@@ -1,9 +1,9 @@
-// Copyright (C) 2024, Benjamin Drung <bdrung@posteo.de>
+// Copyright (C) 2024-2026, Benjamin Drung <bdrung@posteo.de>
 // SPDX-License-Identifier: ISC
 
 use std::env::set_current_dir;
 use std::fs::{create_dir, read_dir, File};
-use std::io::{ErrorKind, Write};
+use std::io::{ErrorKind, Stderr, Write};
 use std::num::NonZeroU32;
 use std::path::Path;
 use std::process::ExitCode;
@@ -12,7 +12,7 @@ use glob::Pattern;
 use lexopt::prelude::*;
 
 use threecpio::examine::examine_cpio_content;
-use threecpio::extract::{extract_cpio_archive, ExtractOptions};
+use threecpio::extract::{extract_cpio_archive, ExtractOptions, ExtractTarget};
 use threecpio::logger::{Level, Logger};
 use threecpio::ranges::Ranges;
 use threecpio::{create_cpio_archive, get_cpio_archive_count, list_cpio_content};
@@ -50,9 +50,30 @@ impl Args {
     }
 }
 
+/// A println function that behaves like tools from the GNU project.
+fn gnu_println(args: std::fmt::Arguments) {
+    let mut stdout = std::io::stdout().lock();
+
+    if let Err(e) = writeln!(stdout, "{}", args) {
+        if e.kind() == std::io::ErrorKind::BrokenPipe {
+            std::process::exit(141); // 128 + SIGPIPE
+        } else {
+            let name = std::option_env!("CARGO_BIN_NAME").unwrap();
+            let _ = writeln!(std::io::stderr(), "{}: stdout write error: {}", name, e);
+            std::process::exit(1);
+        }
+    }
+}
+
+macro_rules! gnu_println {
+    ($($arg:tt)*) => {
+        gnu_println(format_args!($($arg)*))
+    };
+}
+
 fn print_help() {
     let executable = std::env::args().next().unwrap();
-    println!(
+    gnu_println!(
         "Usage:
     {executable} --count ARCHIVE
     {executable} {{-c|--create}} [-v|--debug] [-C DIR] [--data-align BYTES] [ARCHIVE] < manifest
@@ -88,7 +109,7 @@ Optional arguments:
 fn print_version() {
     let name = std::option_env!("CARGO_BIN_NAME").unwrap();
     let version = std::option_env!("CARGO_PKG_VERSION").unwrap();
-    println!("{name} {version}");
+    gnu_println!("{name} {version}");
 }
 
 fn parse_args() -> Result<Args, lexopt::Error> {
@@ -282,6 +303,53 @@ fn print_cpio_archive_count<W: Write>(mut archive: File, out: &mut W) -> std::io
     Ok(())
 }
 
+fn operate_on_archive(
+    archive: File,
+    args: &Args,
+    cwd: &Path,
+    logger: &mut Logger<Stderr>,
+) -> (&'static str, std::io::Result<()>) {
+    let mut stdout = std::io::stdout();
+    if args.count {
+        (
+            "count number of cpio archives",
+            print_cpio_archive_count(archive, &mut stdout),
+        )
+    } else if args.examine {
+        (
+            "examine content",
+            examine_cpio_content(archive, &mut stdout, args.raw),
+        )
+    } else if args.extract {
+        (
+            "extract content",
+            extract_cpio_archive(
+                archive,
+                if args.to_stdout {
+                    ExtractTarget::WritableStream(&mut stdout)
+                } else {
+                    ExtractTarget::Directory(cwd.to_path_buf())
+                },
+                &args.extract_options(),
+                logger,
+            ),
+        )
+    } else if args.list {
+        (
+            "list content",
+            list_cpio_content(
+                archive,
+                &mut stdout,
+                args.parts.as_ref(),
+                &args.patterns,
+                args.log_level.clone(),
+            ),
+        )
+    } else {
+        unreachable!("no operation specified");
+    }
+}
+
 fn main() -> ExitCode {
     let executable = std::env::args().next().unwrap();
     let args = match parse_args() {
@@ -348,49 +416,21 @@ fn main() -> ExitCode {
         }
     }
 
-    let mut stdout = std::io::stdout();
-    let (operation, result) = if args.count {
-        (
-            "count number of cpio archives",
-            print_cpio_archive_count(archive, &mut stdout),
-        )
-    } else if args.examine {
-        (
-            "examine content",
-            examine_cpio_content(archive, &mut stdout, args.raw),
-        )
-    } else if args.extract {
-        (
-            "extract content",
-            extract_cpio_archive(
-                archive,
-                args.to_stdout.then_some(&mut stdout),
-                &args.extract_options(),
-                &mut logger,
-            ),
-        )
-    } else if args.list {
-        (
-            "list content",
-            list_cpio_content(
-                archive,
-                &mut stdout,
-                args.parts.as_ref(),
-                &args.patterns,
-                args.log_level,
-            ),
-        )
-    } else {
-        unreachable!("no operation specified");
+    let cwd = match std::env::current_dir() {
+        Ok(cwd) => cwd,
+        Err(e) => {
+            eprintln!("{executable}: Error: Failed to determine current directory: {e}");
+            return ExitCode::FAILURE;
+        }
     };
-
+    let (operation, result) = operate_on_archive(archive, &args, &cwd, &mut logger);
     if let Err(e) = result {
         match e.kind() {
             ErrorKind::BrokenPipe => {}
             _ => {
                 eprintln!(
                     "{executable}: Error: Failed to {operation} of '{}': {e}",
-                    args.archive.unwrap(),
+                    args.archive.as_deref().unwrap(),
                 );
                 return ExitCode::FAILURE;
             }
