@@ -1,11 +1,13 @@
-// Copyright (C) 2024, Benjamin Drung <bdrung@posteo.de>
+// Copyright (C) 2024-2026, Benjamin Drung <bdrung@posteo.de>
 // SPDX-License-Identifier: ISC
 
 use std::env;
 use std::error::Error;
-use std::fs::{symlink_metadata, File};
+use std::fs::{symlink_metadata, File, OpenOptions};
 use std::io::{ErrorKind, Read, Write};
 use std::os::unix::fs::MetadataExt;
+use std::os::unix::io::{FromRawFd, IntoRawFd};
+use std::os::unix::net::UnixStream;
 use std::process::{Command, Output, Stdio};
 use std::time::SystemTime;
 
@@ -134,6 +136,40 @@ fn test_create_compressed_cpio_file() -> Result<(), Box<dyn Error>> {
     Ok(())
 }
 
+#[test]
+fn test_create_compressed_cpio_file_no_source_date_epoch() -> Result<(), Box<dyn Error>> {
+    let temp_dir = TempDir::new()?;
+    let path = temp_dir.path.join("test.cpio");
+    let path = path.into_os_string().into_string().unwrap();
+
+    let mut cmd = get_command();
+    cmd.args(["--create", &path])
+        .env_remove("SOURCE_DATE_EPOCH")
+        .stdin(Stdio::piped())
+        .stdout(Stdio::piped())
+        .stderr(Stdio::piped());
+    let process = cmd.spawn()?;
+    let mut stdin = process.stdin.as_ref().unwrap();
+    stdin.write_all(b"#cpio: gzip -1\ntests/generate\tgenerate\n")?;
+    process
+        .wait_with_output()?
+        .assert_stdout("")
+        .assert_stderr("");
+
+    let mut cmd = get_command();
+    cmd.arg("-x")
+        .arg("-C")
+        .arg(temp_dir.path.join("extracted"))
+        .arg("-v")
+        .arg(&path);
+    cmd.output()?
+        .assert_stderr("generate\n")
+        .assert_success()
+        .assert_stdout("");
+    assert!(temp_dir.path.join("extracted/generate").exists());
+    Ok(())
+}
+
 #[test]
 fn test_create_compressed_cpio_on_stdout() -> Result<(), Box<dyn Error>> {
     let mut cmd = get_command();
@@ -276,6 +312,44 @@ fn test_create_data_align_zero() -> Result<(), Box<dyn Error>> {
     Ok(())
 }
 
+#[test]
+fn test_create_invalid_manifest() -> Result<(), Box<dyn Error>> {
+    let mut cmd = get_command();
+    cmd.arg("--create")
+        .stdin(Stdio::piped())
+        .stdout(Stdio::piped())
+        .stderr(Stdio::piped());
+    let process = cmd.spawn()?;
+    let mut stdin = process.stdin.as_ref().unwrap();
+    stdin.write_all(b"#cpio: brotli\n")?;
+    process
+        .wait_with_output()?
+        .assert_failure(1)
+        .assert_stderr_contains(
+            "Error: Failed to create 'cpio on stdout': \
+            line 1: Unknown compression format: brotli",
+        )
+        .assert_stdout("");
+    Ok(())
+}
+
+#[test]
+fn test_create_missing_path() -> Result<(), Box<dyn Error>> {
+    let temp_dir = TempDir::new()?;
+    let path = temp_dir.path.join("nonexistent").join("empty.cpio");
+    let path = path.into_os_string().into_string().unwrap();
+
+    let mut cmd = get_command();
+    cmd.args(["--create", &path]);
+    cmd.output()?
+        .assert_failure(1)
+        .assert_stderr_contains(&format!(
+            "Error: Failed to create '{path}': No such file or directory"
+        ))
+        .assert_stdout("");
+    Ok(())
+}
+
 #[test]
 fn test_create_uncompressed_plus_zstd_on_stdout() -> Result<(), Box<dyn Error>> {
     let mut cmd = get_command();
@@ -335,7 +409,7 @@ fn test_examine_compressed_cpio_raw() -> Result<(), Box<dyn Error>> {
             .assert_stderr("")
             .assert_success()
             .assert_stdout(format!(
-                "0\t512\t512\tcpio\t8\n512\t{size}\t{}\t{compression}\t56\n",
+                "0\t512\t512\tcpio\t8\n512\t{size}\t{}\t{compression}\t58\n",
                 size - 512
             ));
     }
@@ -419,10 +493,11 @@ fn test_extract_with_subdir() -> Result<(), Box<dyn Error>> {
 
     println!("tempdir = {:?}", tempdir.path);
     cmd.output()?
-        .assert_stderr(".\npath\npath/file\n.\nusr\nusr/bin\nusr/bin/sh\n")
+        .assert_stderr(".\npath\npath/file\n.\nusr\nusr/bin\nusr/bin/ash\nusr/bin/sh\n")
         .assert_success()
         .assert_stdout("");
     assert!(tempdir.path.join("subdir1/path/file").exists());
+    assert!(tempdir.path.join("subdir2/usr/bin/ash").exists());
     assert!(tempdir.path.join("subdir2/usr/bin/sh").exists());
     Ok(())
 }
@@ -475,11 +550,23 @@ fn test_list_content_compressed_cpio() -> Result<(), Box<dyn Error>> {
         cmd.output()?
             .assert_stderr("")
             .assert_success()
-            .assert_stdout(".\npath\npath/file\n.\nusr\nusr/bin\nusr/bin/sh\n");
+            .assert_stdout(".\npath\npath/file\n.\nusr\nusr/bin\nusr/bin/ash\nusr/bin/sh\n");
     }
     Ok(())
 }
 
+#[test]
+fn test_list_content_compressed_cpio_path_unset() -> Result<(), Box<dyn Error>> {
+    let mut cmd = get_command();
+    cmd.arg("--list").arg("tests/gzip.cpio").env_remove("PATH");
+
+    cmd.output()?
+        .assert_stderr("")
+        .assert_success()
+        .assert_stdout(".\npath\npath/file\n.\nusr\nusr/bin\nusr/bin/ash\nusr/bin/sh\n");
+    Ok(())
+}
+
 #[test]
 fn test_list_content_compressed_cpio_verbose() -> Result<(), Box<dyn Error>> {
     for compression in ["bzip2", "gzip", "lz4", "lzma", "lzop", "xz", "zstd"] {
@@ -500,12 +587,28 @@ fn test_list_content_compressed_cpio_verbose() -> Result<(), Box<dyn Error>> {
                  drwxrwxr-x   2 root     root            0 Apr 14  2024 .\n\
                  drwxrwxr-x   2 root     root            0 Apr 14  2024 usr\n\
                  drwxrwxr-x   2 root     root            0 Apr 14  2024 usr/bin\n\
+                 lrwxrwxrwx   1 root     root            2 Apr 14  2024 usr/bin/ash -> sh\n\
                  -rw-rw-r--   1 root     root           56 Apr 14  2024 usr/bin/sh\n",
             );
     }
     Ok(())
 }
 
+#[test]
+fn test_list_content_invalid_archive() -> Result<(), Box<dyn Error>> {
+    let mut cmd = get_command();
+    cmd.arg("--list").arg("tests/generate");
+
+    cmd.output()?
+        .assert_failure(1)
+        .assert_stderr_contains(
+            "Error: Failed to list content of 'tests/generate': \
+             Failed to determine CPIO or compression magic number",
+        )
+        .assert_stdout("");
+    Ok(())
+}
+
 #[test]
 fn test_list_content_parts_compressed_cpio() -> Result<(), Box<dyn Error>> {
     let mut cmd = get_command();
@@ -590,6 +693,49 @@ fn test_print_version() -> Result<(), Box<dyn Error>> {
     Ok(())
 }
 
+#[test]
+fn test_stdout_write_pipe_fail() -> Result<(), Box<dyn Error>> {
+    for argument in ["--help", "--version"] {
+        let (reader, writer) = UnixStream::pair()?;
+        // Close the reader before spawning (similar to "3cpio --help | true")
+        drop(reader);
+        let stdout = unsafe { Stdio::from_raw_fd(writer.into_raw_fd()) };
+
+        let mut cmd = get_command();
+        cmd.arg(argument)
+            .env("LANG", "C.UTF-8")
+            .stdout(stdout)
+            .stderr(Stdio::piped());
+        let process = cmd.spawn()?;
+
+        process
+            .wait_with_output()?
+            .assert_stderr("")
+            .assert_failure(141);
+    }
+    Ok(())
+}
+
+#[test]
+fn test_stdout_write_pipe_full() -> Result<(), Box<dyn Error>> {
+    for argument in ["--help", "--version"] {
+        let dev_full = OpenOptions::new().write(true).open("/dev/full")?;
+
+        let mut cmd = get_command();
+        cmd.arg(argument)
+            .env("LANG", "C.UTF-8")
+            .stdout(Stdio::from(dev_full))
+            .stderr(Stdio::piped());
+        let process = cmd.spawn()?;
+
+        process
+            .wait_with_output()?
+            .assert_stderr("3cpio: stdout write error: No space left on device (os error 28)\n")
+            .assert_failure(1);
+    }
+    Ok(())
+}
+
 #[test]
 fn test_unexpected_option() -> Result<(), Box<dyn Error>> {
     let mut cmd = get_command();
